// --- [adxl.c]----------------------------------------------------------------
//
// Der ADXL kann alternativ via TWI oder SPI angesprochen werden
// dazu in der global.h die Konstante ADXL_CONNECTION einstellen
//
// Es gibt verschiedene Arbeitsweisen zwischen mc und ADXL vorbereitet:
// - SINGLE
//   jeder Datensatz wird INT1-gesteuert eingelesen und ausgegeben,
//   zustzlich lst das TAP-Ereignis INT2 aus, es wird eine Meldung und die Richtung der Bewegung ausgegeben
// - EVENT 
// - eine Lageabweichung auf der X/Y/Z-Achse lst den INT1 aus,
//   es wird ein akujstischer Alarm ausgegeben solange, bis der Sensor wieder neue stabile Werte erkennt (INT2)
// - STREAM
//   sobald 16 Datenstze den Puffer gefllt haben, wird der Watermark-INT aufgerufen,
//   es werden 16 Datenstze ausgelesen und ausgegeben


#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <stdlib.h>
#include <util/delay.h>
#include "global.h"

#include "adxl.h"
#include "rs232_int.h"
#include "TWI_Master_c.h"
#include "soft_spi_master.h"

#if ADXL_CONNECTION == TWI
uint8_t TWI_Buf[8];								// fr die TWI-Kommunikation
#endif

int16_t s[3];										// fr ein Datensatz Messwerte
int16_t delta[3] = {0,0,0};					// ggf. Korrekturwerte dazu
char msg[10];										// zur Umwandlung von Zahlen in Zeichenfolge
														// Flags fr Interrupts, Counter fr Alarmsummer
volatile uint8_t flag1 = 0, flag2 = 0, tim0_cnt = 0;

uint8_t ADXL_write_register(uint8_t reg, uint8_t value);
uint8_t ADXL_read_register(uint8_t reg);
void    ADXL_read_data(void);
void    ADXL_clear_buffer(void);

void rs232_print_ADXL_DATA(void);
void wait_10ms(uint8_t t);
void alarm_on(void);
void alarm_off(void);

#if ADXL_CONNECTION == TWI

// ----------------------------------------------------------------------------
// schreibt einen Wert in ein Register des ADXL
// prft, ob derselbe Wert auch wieder ausgelesen werden kann
uint8_t ADXL_write_register(uint8_t reg, uint8_t value)
{

TWI_Buf[0] = ADXL_TWI_ADR;
TWI_Buf[1] = reg;
TWI_Buf[2] = value;
TWI_MA_Start_Transceiver_With_Data(TWI_Buf, 3);

if ( ADXL_read_register(reg) == value) return(NO_ERROR); else return (ERROR);
}

// ----------------------------------------------------------------------------
// liest den Inhalt eines Registers
uint8_t ADXL_read_register(uint8_t reg)
{
TWI_Buf[0] = ADXL_TWI_ADR;
TWI_Buf[1] = reg;
TWI_MA_Start_Transceiver_With_Data(TWI_Buf, 2);

TWI_Buf[0] = ADXL_TWI_ADR + 1;
TWI_MA_Start_Transceiver_With_Data(TWI_Buf, 2);	// Daten via TWI abholen
TWI_MA_Get_Data_From_Transceiver( TWI_Buf, 2);	// Daten umkopieren

return (TWI_Buf[1]);								// das 2.Element enthlt den Wert
}

// ----------------------------------------------------------------------------
// liest einen Datensatz von 3 * 2 Byte mit den x, y, z-Werten
// wandelt die Byte-Paare jeweils in int16-Werte um
void ADXL_read_data(void)
{
uint8_t i, j = 0;

TWI_Buf[0] = ADXL_TWI_ADR;
TWI_Buf[1] = ADXL_DATA;
TWI_MA_Start_Transceiver_With_Data(TWI_Buf, 2);

TWI_Buf[0] = ADXL_TWI_ADR + 1;
TWI_MA_Start_Transceiver_With_Data(TWI_Buf, 7);	// Daten via TWI abholen
TWI_MA_Get_Data_From_Transceiver( TWI_Buf, 7);	// Daten umkopieren

for (i = 1; i < 7; i = i + 2)
	{
	s[j]  = TWI_Buf[i + 1] * 256;				// eingelesenen Wert in int16 umwandeln
	s[j] += TWI_Buf[i];
	s[j] -= delta[j];
	j++;
	}
}

#elif ADXL_CONNECTION == SPI
// ----------------------------------------------------------------------------
// schreibt einen Wert in genau ein Register des ADXL
// prft, ob derselbe Wert auch wieder ausgelesen werden kann
uint8_t ADXL_write_register(uint8_t reg, uint8_t value)
{
uint8_t a[2];

reg &= 0x3F;										// obere beiden Bits maskieren

#ifdef DEBUG
put_s("0x");
put_s(utoa(reg, msg, 16));
put_s(" -> ");
put_s(utoa(value, msg,16));
#endif

// das Write-Bit lschen und das Multiple_Byte Bit setzen
// und die Registeradresse (die < 32 ist !) addieren
//a[0] = 0x01000000 + reg;
a[0] = 0x40 + reg;
a[1] = value;										// der Wert des Register

ADXL_SPI_write_array(a, 2);					// das Array schreiben

a[0] = 0xC0 + reg;								// Zum Test das geschriebene Byte 
ADXL_SPI_write_array(a, 2);					// neu einlesen und vergleichen

#ifdef DEBUG
if (a[1] == value) put_s(" ok"); else put_s(" err");
put_c(13);
#endif

if (a[1] == value) return(NO_ERROR); else return(ERROR);

}

// ----------------------------------------------------------------------------
// liest den Inhalt genau eines Registers des ADXL
uint8_t ADXL_read_register(uint8_t reg)
{
uint8_t a[2];
// das Write-Bit und das Multiple_Byte Bit setzen
// und die Registeradresse (die < 64 sein muss !) addieren

a[0] = 0xC0 + (reg & 0x3F);
ADXL_SPI_write_array(a, 2);

return(a[1]);										// das gelesene Byte steht in a[1]

}

// ----------------------------------------------------------------------------
// liest einen Datensatz von 3 * 2 Byte mit den x, y, z-Werten
// wandelt die Byte-Paare jeweils in int16-Werte um
// berechnet die Quadrate der einzelnen Vektorkomponenten und
// ermittelt die Lnge des Vektors als Summe der Quadrate der Komponenten
// und zieht davon das Quadrat des Ruhewertes 1g ab.
void ADXL_read_data(void)
{
uint8_t i,j;
uint8_t a[7];

// das Write-Bit und das Multiple_Byte Bit setzen
// und die Registeradresse (die < 64 sein muss !) addieren
a[0] = (0xC0 + ADXL_DATA);
ADXL_SPI_write_array(a, 7);					// 6 Byte ab ADXL_DATA holen

j = 0;

for (i = 1; i < 7; i = i + 2)
	{
	s[j]  = a[i + 1] * 256;						// das High-Byte schieben
	s[j] += a[i];									// das Low-Byte addieren
	s[j] -= delta[j];								// ggf. Korrekturwert bercksichtigen
	j++;
	}
}

#endif

/* - die wesentlichen Register fr die Init-Sequenzen  ...
uint8_t err = 0;
if (! (ADXL_read_register(ADXL_DEVID) == ADXL_ID)) err = 128;
err += ADXL_write_register(ADXL_THRESH_TAP,  0x5);		// 20 * 62.5mg
err += ADXL_write_register(ADXL_TAP_DUR,     0xAF);
err += ADXL_write_register(ADXL_TAP_AXES,   0x06);
err += ADXL_write_register(ADXL_INT_ENABLE,  0x40);		// DATAReady, Activity, Inactivity enabled
err += ADXL_write_register(ADXL_INT_MAP,  0x40);			// DATAready = INT1, Activity = INT2, Inactivity = INT2
err += ADXL_write_register(ADXL_DATA_FORMAT, DATA_INT_INVERT | DATA_2G);
err += ADXL_write_register(ADXL_FIFO_CTL,    FIFO_CTL_FIFO_BYPASS);
err += ADXL_write_register(ADXL_POWER_CTL,   POWER_CTL_LINK | POWER_CTL_MEASURE);

am Ende muss immer err==0 sein, sonst ist ein Fehler aufgetreten
*/

#if ADXLMODE == SINGLE
// ----------------------------------------------------------------------------
// Initialisiert den ADXL345 fr TAP-Interrupt in X/Y Richtung
// Ein Interrupt wird ausgelst, wenn der eingestellte Schwellwert THRESH_TAP berschritten wird
// in einem Zeitraum, der krzer als THRESH_DUR ist.
// Reagiert auf Bewegung in X/Y-Richtung mit der Meldung "Single_TAP" + Angabe der
// erkannten Bewegungs-Richtung
uint8_t ADXL_init(void)
{
uint8_t err = 0;

// der Chip meldet sich mit einer festen ID,  wenn nicht, dann Fehler
if (!(ADXL_read_register(ADXL_DEVID) == ADXL_ID)) err = 128;

err += ADXL_write_register(ADXL_DATA_FORMAT, DATA_INT_INVERT | DATA_2G);
err += ADXL_write_register(ADXL_BW_RATE,     BW_RATE_6_25);
err += ADXL_write_register(ADXL_FIFO_CTL,    FIFO_CTL_FIFO_BYPASS); // keine Daten puffern

err += ADXL_write_register(ADXL_THRESH_TAP,  0x10);	// 10 * 62.5mg
err += ADXL_write_register(ADXL_TAP_DUR,     0x80);	// 128 * 600us
err += ADXL_write_register(ADXL_TAP_AXES,    0x06);	// ohne Z-Achse

err += ADXL_write_register(ADXL_ACT_INACT_CTL,0x00);	// deaktivieren

err += ADXL_write_register(ADXL_INT_ENABLE,  0xC0);	// DATA_READY + SINGLE_TAP
err += ADXL_write_register(ADXL_INT_MAP,  	0x40);	// SINGLE_TAP nach INT2

err += ADXL_write_register(ADXL_POWER_CTL,    POWER_CTL_LINK | POWER_CTL_MEASURE);

return(err);
}

#elif ADXLMODE == EVENT
// ----------------------------------------------------------------------------
// Initialisiert den ADXL345 fr Interrupt-Modus
// Daten werden nicht ausgelesen
// Activity- und Inactivity-Interrupt werden genutzt.
// Bei Bewegung des Sensors schaltet der Interrupt
// Activity einen Alarm (Summer) an, 
// nach Ablauf der Zeit TIME_INACT schaltet der Interrupt Inactivity ihn wieder ab
uint8_t ADXL_init(void)
{
uint8_t err = 0;

// der Chip meldet sich mit einer festen ID,  wenn nicht, dann Fehler
if (!(ADXL_read_register(ADXL_DEVID) == ADXL_ID)) err = 128;


//err += ADXL_write_register(ADXL_TRESH_TAP,   250);		// 20 * 62.5mg
//err += ADXL_write_register(ADXL_TAP_AXES, 0x00);
err += ADXL_write_register(ADXL_DATA_FORMAT, DATA_INT_INVERT | DATA_2G);

err += ADXL_write_register(ADXL_THRESH_ACT,   10);		// 10 * 62.5mg
err += ADXL_write_register(ADXL_THRESH_INACT, 10);	   // 10 * 62.5mg
err += ADXL_write_register(ADXL_TIME_INACT,   5);		// 5 * 1sec
err += ADXL_write_register(ADXL_ACT_INACT_CTL, 0xFF); //  AC-coupled, alle Achsen aktiv
err += ADXL_write_register(ADXL_INT_MAP,       0x08);	// Activity = INT0, Inactivity = INT1
err += ADXL_write_register(ADXL_INT_ENABLE,    0x18);	// Activity, Inactivity enablen

err += ADXL_write_register(ADXL_BW_RATE,   BW_RATE_25);
err += ADXL_write_register(ADXL_FIFO_CTL,  FIFO_CTL_FIFO_BYPASS); // keine Daten puffern
err += ADXL_write_register(ADXL_POWER_CTL, POWER_CTL_LINK | POWER_CTL_MEASURE);

return(err);
}

#elif ADXLMODE == STREAM
// ----------------------------------------------------------------------------
// Initialisiert den ADXL345 fr den Stream-Modus,
// der Watermark-Interrupt veranlasst das Einlesen eines Blocks von Daten
// Gibt die Daten in 16er Blcken aus (immer dann, wenn der Watermark-INT gesetzt wird)

uint8_t ADXL_init(void)
{
uint8_t err = 0;

// der Chip meldet sich mit einer festen ID,  wenn nicht, dann Fehler
if (!(ADXL_read_register(ADXL_DEVID) == ADXL_ID)) err = 128;

err += ADXL_write_register(ADXL_DATA_FORMAT, DATA_INT_INVERT | DATA_2G);

err += ADXL_write_register(ADXL_INT_MAP,  0x00);			// Watermark auf INT1
err += ADXL_write_register(ADXL_INT_ENABLE, INT_WATERMARK); // INT_Watermark enablen

err += ADXL_write_register(ADXL_BW_RATE,   BW_RATE_12_5);
err += ADXL_write_register(ADXL_FIFO_CTL,  FIFO_CTL_FIFO_STREAM + ADXL_BLOCK_SIZE);
err += ADXL_write_register(ADXL_POWER_CTL, POWER_CTL_LINK | POWER_CTL_MEASURE);

return(err);
}

#endif

// ----------------------------------------------------------------------------
// Beim Start zeigen die Messwerte auch bei einem horizontal ausgerichteten Sensor
// nicht genau auf die erwarteten Werte 0,0,255.
// Hier werden Korrekturwerte berechnet (und optional in die Register des ADXL geschrieben).
// Durch die Korrekturwerte soll erreicht werden, dass in der Ausgangslage
// die X/Y Werte zu 0 tendieren und der Z-Wert sich 255 annhert.
// Es wird der Mittelwert aus 16 Einzelmessungen gebildet.
// Der Sensor darf in der Kalibrierzeit nicht bewegt werden.

// Die Korrekturwerte knnten auch in die Offset-Register des ADXL geschrieben
// werden, allerdings kann dabei nur 1/4 der Auflsung (16mg) gearbeitet werden.
// Manuell lt sich die volle Auflsung (4mg) erreichen.

// Sinn macht die Aktion etwa, wenn der ADXL als Wasserwaage o.. genutzt werden soll

void ADXL_vertikal_calibrate(void)
{
uint8_t j;
int16_t x[3] = {0,0,0};

j = 16;
while(j--)											// es wird ber 16 Messungen der Mittelwert gebildet
	{
	ADXL_read_data();
	x[0] += s[0];
	x[1] += s[1];
	x[2] += s[2];
	wait_10ms(15);									// hinreichend groe Wartezeit fr geringe Datenraten einlegen
	}

delta[0] = x[0] / 16;							// diese Werte werden fr die manuelle Korrektur
delta[1] = x[1] / 16;							// verwendet
delta[2] = x[2] / 16;
delta[2] -= 255;									// der Normalwert fr die Z_Achse ist 1g = 255

put_s("Korrekturwerte X/Y/Z : ");
put_s(ltoa(delta[0],msg, 10));
put_c(',');
put_s(ltoa(delta[1],msg, 10));
put_c(',');
put_s(ltoa(delta[2],msg, 10));
put_c(13);

/*
x[0] /= 4;											// diese Korrekturwerte werden in die Offset-
x[0]  = 0 - x[0];									// register geschrieben,
														// LSB entspricht 15.6mg = 4x Normalauflsung von 4mg
x[1] /= 4;
x[1]  = 0 - x[1];

x[2]  = 0x00FF - x[2];
x[2] /= 4;

put_s("Korrekturwerte: ");
put_s(ltoa(x[0],msg, 10));
put_c(',');
put_s(ltoa(x[1],msg, 10));
put_c(',');
put_s(ltoa(x[2],msg, 10));
put_c(13);

ADXL_write_register(ADXL_OFSX, (int8_t) x[0]);
ADXL_write_register(ADXL_OFSY, (int8_t) x[1]);
ADXL_write_register(ADXL_OFSZ, (int8_t) x[2]);
*/

}

// ----------------------------------------------------------------------------
// Liest alle 32 Datenstze ein, damit alle Interruptflags gelscht werden fr die
// Ereignisse: dataready, watermark, overflow.
// Zum Lschen der anderen Interrupts muss das Register INT_SOURCE gelesen werden !

void ADXL_clear_buffer(void)
{
uint8_t i = 32;
while(i--)	ADXL_read_data();

}

// ----------------------------------------------------------------------------
// gibt zum Debuggen alle Register des ADXL345 mit Adresse und aktuellem Wert aus
#ifdef DEBUG
void ADXL_print_registers(void)
{
uint8_t i;
for (i = ADXL_DEVID; i < 0x3A; i++)
	{
	put_s("0x");
	put_s(itoa(i, msg, 16));
	put_s(" = 0x");
	put_s(itoa(ADXL_read_register(i), msg, 16));
	put_c(13);
	
	if (i == ADXL_DEVID) i = (0x1d - 1);	// die Lcke zwischen ADXL_DEVID und 0x1D berspringen
	}
}
#endif

// ----------------------------------------------------------------------------
int main( void )
{
uint8_t i, j; 

#if ADXL_CONNECTION == TWI
TWI_Master_Initialise();						// keine Parmeter erforderlich

#elif ADXL_CONNECTION == SPI
SPI_master_init();
#elif

error
#endif

sei();												// Interrupt fr TWI notwendig !!

rs232_init();										// auf 9600 Baud

/*
put_s("Hello world");							// zeigt, dass die serielle Schnittstelle lebt
put_c(13);
wait_10ms(50);
*/

#ifdef DEBUG
put_s("ADXL ID: ");
put_s(utoa(ADXL_read_register(0), msg, 16));
put_c(13);


// Die Registerwerte vor der Initialisierung ausgeben
put_s("Register vor Init:");
put_c(13);
ADXL_print_registers();
put_c(13);
put_s("writing registers:");
put_c(13);
#endif

i = ADXL_init();									// hier werden die Register beschrieben
														// dabei wird protokolliert
put_c(13);
put_s("ADXL ");
if (i)												// falls i > 0, dann gab es Fehler
	{
	put_s("error: ");
	put_s(utoa(i, msg, 10));
	}
else
	{
	put_s("ok");
	}
put_c(13);

#ifdef DEBUG
// die Registerwerte nach der Initialisierung ausgeben
put_s("Register nach Init");
put_c(13);
ADXL_print_registers();
#endif

//ADXL_write_register(ADXL_POWER_CTL,  POWER_CTL_LINK | POWER_CTL_MEASURE);
//ADXL_vertikal_calibrate();						// versucht den ADXL auf 0,0,255 zu justieren


ADXL_clear_buffer();								// damit werden alle aufgelaufenen Interrupts freigegeben
ADXL_read_register(ADXL_INT_SOURCE); 		// und hier alle Flags gelscht werden
														// INT0/INT1 am uc fr INT1/INT2 am ADXL konfigurieren
EICRA = (1<< ISC11 | 0<<ISC10 | 1<< ISC01 | 0<<ISC00); // falling edge
EIMSK = (1<< INT1  | 1<<INT0);				// INT0 und INT1 aktivieren

while(1)
	{

#if ADXLMODE == SINGLE
	if (flag1)
		{
		flag1 = 0;
		ADXL_read_register(ADXL_INT_SOURCE); // Flags lschen
		ADXL_read_data();							// einen Datensatz einlesen
		put_s("INT1 : ");
		rs232_print_ADXL_DATA();				// zur Ausgabe auf rs232 weiterreichen
		}

	if (flag2)										// flag2 wird von INT1 gesetzt
		{
		flag2= 0;

		j = ADXL_read_register(ADXL_ACT_TAP_STATUS);
//		put_s("TAP_Status: ");
//		put_s(itoa(j, msg, 16));
//		put_c(13);
		
		i = ADXL_read_register(ADXL_INT_SOURCE); // Flags lschen		
		put_s("INT2 : ");		
//		put_s(itoa(i, msg, 16));

		if (i & INT_ACTIVITY)	put_s("ACTIV ");
		if (i & INT_INACTIVITY)	put_s("INACTIV ");
		if (i & INT_SINGLE_TAP)	put_s("Single_TAP -> ");
		if (i & INT_DOUBLE_TAP)	put_s("Double_TAP ");

		if (j & 0x04) put_c('X');					// die Auslse-Richtung
		else if(j & 0x02) put_c('Y');
		
		put_c(13);
		}

#elif ADXLMODE == EVENT
		if (flag2)										// flag1/2 wird von INT0/1 gesetzt
			{
			flag2= 0;
			i = ADXL_read_register(ADXL_INT_SOURCE);
			put_s("INT2: ") ;
			put_s(itoa(i, msg, 16));
			put_c(13);
			ADXL_read_data();							// einen Datensatz einlesen
			rs232_print_ADXL_DATA();				// zur Ausgabe auf rs232 weiterreichen
			
			alarm_off();
			}

		if (flag1)
			{
			flag1 = 0;
			i = ADXL_read_register(ADXL_INT_SOURCE);
			put_s("INT1: ") ;
			put_s(itoa(i, msg, 16));
			put_c(13);
			ADXL_read_data();							// einen Datensatz einlesen
			rs232_print_ADXL_DATA();				// zur Ausgabe auf rs232 weiterreichen
			
			alarm_on();
			}

#endif


#if ADXLMODE == STREAM
		if (flag2)								// flag1/2 wird von INT0/1 gesetzt
			{
			flag2= 0;
			i = ADXL_read_register(ADXL_INT_SOURCE);
			put_s("INT2: ") ;
			put_s(itoa(i, msg, 16));
			put_c(13);
			}

		if (flag1)
			{
			flag1 = 0;
//			i = ADXL_read_register(ADXL_INT_SOURCE);
//			put_s("INT1: ") ;
//			put_s(itoa(i, msg, 30));
//			i = 0;
//			put_c(13);

			i = ADXL_BLOCK_SIZE;							// Anzahl der Datenstze fr Watermark-INT
			while(i--)
				{
				ADXL_read_data();							// einen Datensatz einlesen
				rs232_print_ADXL_DATA();				// zur Ausgabe auf rs232 weiterreichen
				}

//			i = ADXL_read_register(ADXL_INT_SOURCE);
//			put_s("INT1: ") ;
//			put_s(itoa(i, msg, 30));
//			put_c(13);

			}
#endif
	}
}


// ----------------------------------------------------------------------------
// Setzt die notwendigen TimerRegister des mc fr einen akustischen Alarm
// der Summer ist an PD6 (OC0A) angeschlossen 
void alarm_on(void)
{

TCCR0A = (1<<COM0A0 | 1<<COM0B0 | 1<< WGM01);  // toogle on Compare Match, CTC-Modus
TCCR0B = (1<<CS01);								// Vorteiler 8, Takt 3.68 MHz CompareA 64 -> 3600 Hz
OCR0A = TONE_HIGH;
DDRD |= (1<<PD6);									// hier ist der Summer angeschlossen !

TIMSK0 = 1<<OCIE0A;								// Compare-Interrupt fr Tonhhenwechsel einspannen
tim0_cnt = 0;
}

// ----------------------------------------------------------------------------
// lscht den Alarm
void alarm_off(void)
{
TCCR0B = 0;											// Timer stoppen
DDRD &=(~(1<<PD6 ));								// Ausgangspin als Input
}

// ----------------------------------------------------------------------------
// jeweils nur ein Flag incrementieren (von 0 auf 1)
ISR (INT0_vect)
{
flag1++;
}

// ----------------------------------------------------------------------------
// jeweils nur ein Flag incrementieren (von 0 auf 1)
ISR (INT1_vect)
{
flag2++;
}

// ----------------------------------------------------------------------------
// nutzt den TIMER0_COMP-Interrupt als Zeitglied, um einen Wechsel
// der Tonfrequenz zu erreichen
ISR (TIMER0_COMPA_vect)
{
register uint8_t t;

t = tim0_cnt;						// jeweils nach berlauf auf 0
t++;									// wird die Tonhhe gendert

if (!(t))							// Der Counter steht auf 0
	{
	if (OCR0A == TONE_HIGH) OCR0A = TONE_LOW; else OCR0A = TONE_HIGH;
	}
tim0_cnt = t;

}

// ----------------------------------------------------------------------------
// wartet jeweils 10ms
void wait_10ms(uint8_t t)
{
while (t--) _delay_ms(10);
}

// ----------------------------------------------------------------------------
// gibt die in s[] zwischengespeicherten Einzelwerte via RS232 aus
void rs232_print_ADXL_DATA(void)
{
uint8_t i;

for (i = 0; i < 3; i++)
	{
	put_s(ltoa(s[i], msg, 10));
	put_c(',');
	}
put_c(13);

}

// --- [eof] ------------------------------------------------------------------