// --- [ soft_spi_master.c ] --------------------------------------------------
// Software SPI, SPI-Mode 3 (CPOL = 1, CPHA = 1)
// Ruhepegel = high, Setup at falling edge, sample at rising edge

// DORD kann unten eingestellt werden

#include <avr/io.h>
#include <util/delay.h>
#include "global.h"

#if ADXL_CONNECTION == SPI
#if SPI_MODE == SOFTWARE

#define SPI_PORT 			PORTC		// PORT fr SPI
#define SPI_PIN			PINC		// PIN  fr SPI
#define SPI_DDR			DDRC		// DDR  fr SPI
#define SPI_SEL			PC5		// Selectpin
#define SPI_CLK 			PC4		// Clockpin
#define SPI_MOSI			PC3		// MasterOut -> Ausgang
#define SPI_MISO			PC2		// MasterIn  -> Eingang

#define SPI_DORD			0			// 0 -> MSB zuerst senden, 1 -> LSB zuerst senden
#define SPI_WAIT			1			// Wartezeit in usec fr 1/2 Takt 
											// Takt dauert 2*5us -> 100KHz
											// funktioniert aber auch mit 1 -> 500KHz


// --- lokale Funktionen ------------------------------------------------------

void spi_wait(void);					// Warteschleife fr 1/2 Takt
uint8_t spi_send(uint8_t);

// ----------------------------------------------------------------------------
// schaltet die Aus-/Eingnge fr eine Software-SPI Kommunikation als Master
void SPI_master_init(void)
{											// Ausgnge schalten
SPI_DDR  |=  (1<<SPI_MOSI | 1<<SPI_SEL | 1<<SPI_CLK);
SPI_PORT |=  (1<<SPI_MOSI | 1<<SPI_SEL | 1<<SPI_CLK); // auf high ziehen
SPI_DDR  &= ~(1<<SPI_MISO);		// Eingang schalten
}

/*
// ----------------------------------------------------------------------------
// bereitet das Senden von Bytes vor, setzt SlaveSelect auf low
// ruft Unterprogramm zum senden des Bytes auf
// und zieht abschlieend alle Pin auf high
// (das lesen ist noch nicht getestet) 

uint8_t SPI_byte_out(uint8_t spi_out)
{
register uint8_t i;
											// alle Ausgnge liegen auf high
SPI_PORT &= ~(1<<SPI_SEL);			// Select auf low legen
spi_wait();								// 1/2 Takt warten							

i = spi_send(spi_out);
											// alle Pins wieder auf high
SPI_PORT |= (1<<SPI_MOSI | 1<< SPI_SEL | 1<<SPI_CLK);					

return(i);
}
*/

// ----------------------------------------------------------------------------
// Schreibt und liest gleichzeitig ein Array per SPI.
// Abhngig von den oberen beiden Bits im 1. Byte liefert der ADXL Daten zurck, 
// deren Startadresse im ersten Byte definiert wurde.

// j ist die Anzahl der zu versendenden/lesenden Bytes (incl. Adress-Byte)
// die Rckgabe steht in data[] (nicht getestet)

void ADXL_SPI_write_array(uint8_t *data, uint8_t j)
{
register uint8_t i;
											// alle Ausgnge liegen auf high
SPI_PORT &= ~(1<<SPI_SEL);			// Select auf low legen
spi_wait();								// 1/2 Takt warten

for (i = 0; i < j; i++)
	{
	data[i] = spi_send(data[i]);
	spi_wait();
	spi_wait();							// einen Takt warten
	}
	
SPI_PORT |= (1<<SPI_MOSI | 1<< SPI_SEL | 1<<SPI_CLK);	

}

// ----------------------------------------------------------------------------
// Sendet die Bits eines Bytes,
// wird sowohl von spi_byte_out() als auch von spi_array_out() genutzt
uint8_t spi_send(uint8_t spi_out)
{
register uint8_t i;
register uint8_t spi_in = 0;

for (i = 0; i < 8; i++)	
	{										
	SPI_PORT &= ~(1<<SPI_CLK);		// Clock auf low
											// Datenbyte setzen whrend clock low ist
#if SPI_DORD == 1
	if (spi_out & 0x01)	 			// wenn Bit.0 gesetzt
#else
	if (spi_out & 0x80)	 			// wenn Bit.7 gesetzt
#endif
		{
		SPI_PORT |= (1<<SPI_MOSI);	// dann MOSI auf high
		}
	else
		{
		SPI_PORT &= ~(1<<SPI_MOSI);// sonst MOSI auf low
		}
#if SPI_DORD == 1
	spi_out = spi_out >> 1;			// das Ausgabebyte um 1 nach rechts schieben	
#else
	spi_out = spi_out << 1;			// das Ausgabebyte um 1 nach links schieben	
#endif	
	spi_wait();							// einen halben Takt warten

	SPI_PORT |= (1<<SPI_CLK);		// Clock auf high

#if SPI_DORD == 1
	spi_in  = spi_in  >> 1;			// das Empfangsbyte nach rechts schieben
#else
	spi_in  = spi_in  << 1;			// das Empfangsbyte nach links schieben
#endif
	
	if (SPI_PIN & 1<<SPI_MISO)		// Datenbit ber MISO einlesen
		{
#if SPI_DORD == 1
		spi_in |= 0x80;				// Bit.7 setzen
#else
		spi_in |= 0x01;				// Bit.0 setzen
#endif
		}

	spi_wait();							// einen halben Takt warten

	}										// das Ganze nun noch 7x je Byte

											// alle Pins ausser SEL wieder auf high
SPI_PORT |= (1<<SPI_MOSI |  1<<SPI_CLK);	

return(spi_in);


}

// ----------------------------------------------------------------------------
// 1/2 Takt warten
void spi_wait(void)
{
_delay_us(SPI_WAIT);
}

#endif
#endif
// --- [ eof ] ----------------------------------------------------------------
